"""
# author: Swain Subrat Kumar
A demo __init__.py file should contain all 
the public API that need to be exported 
from this package
"""

from .awesome_ids_model import AwesomeIDS

__all__ = [
    "AwesomeIDS"
]
