# from .kitsune import KitModel, KeyedKitModel
from .awesome_ids import awesome_ids_model
# do we have to make any changes here to create the function inside awesome ids , I dont want to call kitsune and keyed kitsune here


__all__ = [
    "AwesomeIDS"
    ]